@@ -15,11 +15,11 @@
 # limitations under the License.
 
 import os
-import ssl
 import fcntl
 import errno
 import struct
 import select as builtin_select
+import six
 
 
 def set_blocking(fd, blocking=True):
@@ -32,7 +32,7 @@ def set_blocking(fd, blocking=True):
     old_flag = fcntl.fcntl(fd, fcntl.F_GETFL)
 
     if blocking:
-        new_flag = old_flag &~ os.O_NONBLOCK
+        new_flag = old_flag & ~ os.O_NONBLOCK
     else:
         new_flag = old_flag | os.O_NONBLOCK
 
@@ -41,14 +41,14 @@ def set_blocking(fd, blocking=True):
     return not bool(old_flag & os.O_NONBLOCK)
 
 
-def select(read_streams, timeout=0):
+def select(read_streams, write_streams, timeout=0):
     """
-    Select the streams from `read_streams` that are ready for reading.
+    Select the streams from `read_streams` that are ready for reading, and
+    streams from `write_streams` ready for writing.
 
-    Uses `select.select()` internally but returns a flat list of streams.
+    Uses `select.select()` internally but only returns two lists of ready streams.
     """
 
-    write_streams = []
     exception_streams = []
 
     try:
@@ -57,11 +57,12 @@ def select(read_streams, timeout=0):
             write_streams,
             exception_streams,
             timeout,
-        )[0]
+        )[0:2]
     except builtin_select.error as e:
         # POSIX signals interrupt select()
-        if e[0] == errno.EINTR:
-            return []
+        no = e.errno if six.PY3 else e[0]
+        if no == errno.EINTR:
+            return ([], [])
         else:
             raise e
 
@@ -74,7 +75,6 @@ class Stream(object):
     add consistency to the reading of sockets and files alike.
     """
 
-
     """
     Recoverable IO/OS Errors.
     """
@@ -84,7 +84,6 @@ class Stream(object):
         errno.EWOULDBLOCK,
     ]
 
-
     def __init__(self, fd):
         """
         Initialize the Stream for the file descriptor `fd`.
@@ -92,7 +91,9 @@ class Stream(object):
         The `fd` object must have a `fileno()` method.
         """
         self.fd = fd
-
+        self.buffer = b''
+        self.close_requested = False
+        self.closed = False
 
     def fileno(self):
         """
@@ -101,7 +102,6 @@ class Stream(object):
 
         return self.fd.fileno()
 
-
     def set_blocking(self, value):
         if hasattr(self.fd, 'setblocking'):
             self.fd.setblocking(value)
@@ -109,40 +109,82 @@ class Stream(object):
         else:
             return set_blocking(self.fd, value)
 
-
     def read(self, n=4096):
         """
         Return `n` bytes of data from the Stream, or None at end of stream.
         """
 
-        try:
-            if hasattr(self.fd, 'recv'):
-                return self.fd.recv(n)
-            return os.read(self.fd.fileno(), n)
-        except EnvironmentError as e:
-            if e.errno not in Stream.ERRNO_RECOVERABLE:
-                raise e
+        while True:
+            try:
+                if hasattr(self.fd, 'recv'):
+                    return self.fd.recv(n)
+                return os.read(self.fd.fileno(), n)
+            except EnvironmentError as e:
+                if e.errno not in Stream.ERRNO_RECOVERABLE:
+                    raise e
 
 
     def write(self, data):
         """
-        Write `data` to the Stream.
+        Write `data` to the Stream. Not all data may be written right away.
+        Use select to find when the stream is writeable, and call do_write()
+        to flush the internal buffer.
         """
 
         if not data:
             return None
 
+        self.buffer += data
+        self.do_write()
+
+        return len(data)
+
+    def do_write(self):
+        """
+        Flushes as much pending data from the internal write buffer as possible.
+        """
         while True:
             try:
-                if hasattr(self.fd, 'send'):
-                    self.fd.send(data)
-                    return len(data)
-                os.write(self.fd.fileno(), data)
-                return len(data)
+                written = 0
+
+                socket = self.fd
+                if hasattr(socket, 'connection'):
+                    socket = socket.connection
+
+                if hasattr(socket, 'send'):
+                    written = socket.send(self.buffer)
+                else:
+                    written = os.write(socket.fileno(), self.buffer)
+
+                self.buffer = self.buffer[written:]
+
+                # try to close after writes if a close was requested
+                if self.close_requested and len(self.buffer) == 0:
+                    self.close()
+
+                return written
             except EnvironmentError as e:
                 if e.errno not in Stream.ERRNO_RECOVERABLE:
                     raise e
 
+    def needs_write(self):
+        """
+        Returns True if the stream has data waiting to be written.
+        """
+        return len(self.buffer) > 0
+
+    def close(self):
+        self.close_requested = True
+
+        # We don't close the fd immediately, as there may still be data pending
+        # to write.
+        if not self.closed and len(self.buffer) == 0:
+            self.closed = True
+            if hasattr(self.fd, 'close'):
+                self.fd.close()
+            else:
+                os.close(self.fd.fileno())
+
     def __repr__(self):
         return "{cls}({fd})".format(cls=type(self).__name__, fd=self.fd)
 
@@ -171,7 +213,6 @@ class Demuxer(object):
         self.stream = stream
         self.remain = 0
 
-
     def fileno(self):
         """
         Returns the fileno() of the underlying Stream.
@@ -181,11 +222,9 @@ class Demuxer(object):
 
         return self.stream.fileno()
 
-
     def set_blocking(self, value):
         return self.stream.set_blocking(value)
 
-
     def read(self, n=4096):
         """
         Read up to `n` bytes of data from the Stream, after demuxing.
@@ -202,24 +241,15 @@ class Demuxer(object):
         if size <= 0:
             return
         else:
-            data = ''
+            data = six.binary_type()
             while len(data) < size:
-                while True:
-                    try:
-                        nxt = self.stream.read(size - len(data))
-                        if nxt is None:
-                            continue
-                        break
-                    except ssl.SSLError as e:
-                        if e.errno != 2:
-                            raise
+                nxt = self.stream.read(size - len(data))
                 if not nxt:
                     # the stream has closed, return what data we got
                     return data
-                data = "{0}{1}".format(data, nxt)
+                data = data + nxt
             return data
 
-
     def write(self, data):
         """
         Delegates the the underlying Stream.
@@ -227,6 +257,32 @@ class Demuxer(object):
 
         return self.stream.write(data)
 
+    def needs_write(self):
+        """
+        Delegates to underlying Stream.
+        """
+
+        if hasattr(self.stream, 'needs_write'):
+            return self.stream.needs_write()
+
+        return False
+
+    def do_write(self):
+        """
+        Delegates to underlying Stream.
+        """
+
+        if hasattr(self.stream, 'do_write'):
+            return self.stream.do_write()
+
+        return False
+
+    def close(self):
+        """
+        Delegates to underlying Stream.
+        """
+
+        return self.stream.close()
 
     def _next_packet_size(self, n=0):
         size = 0
@@ -235,21 +291,13 @@ class Demuxer(object):
             size = min(n, self.remain)
             self.remain -= size
         else:
-            data = ''
+            data = six.binary_type()
             while len(data) < 8:
-                while True:
-                    try:
-                        nxt = self.stream.read(8 - len(data))
-                        if nxt is None:
-                            continue
-                        break
-                    except ssl.SSLError as e:
-                        if e.errno != 2:
-                            raise
+                nxt = self.stream.read(8 - len(data))
                 if not nxt:
                     # The stream has closed, there's nothing more to read
                     return 0
-                data = "{0}{1}".format(data, nxt)
+                data = data + nxt
 
             if data is None:
                 return 0
@@ -279,14 +327,23 @@ class Pump(object):
     Pumps are selectable based on the 'read' end of the pipe.
     """
 
-    def __init__(self, from_stream, to_stream):
+    def __init__(self,
+                 from_stream,
+                 to_stream,
+                 wait_for_output=True,
+                 propagate_close=True):
         """
         Initialize a Pump with a Stream to read from and another to write to.
+
+        `wait_for_output` is a flag that says that we need to wait for EOF
+        on the from_stream in order to consider this pump as "done".
         """
 
         self.from_stream = from_stream
         self.to_stream = to_stream
-
+        self.eof = False
+        self.wait_for_output = wait_for_output
+        self.propagate_close = propagate_close
 
     def fileno(self):
         """
@@ -297,11 +354,9 @@ class Pump(object):
 
         return self.from_stream.fileno()
 
-
     def set_blocking(self, value):
         return self.from_stream.set_blocking(value)
 
-
     def flush(self, n=4096):
         """
         Flush `n` bytes of data from the reader Stream to the writer Stream.
@@ -313,11 +368,29 @@ class Pump(object):
         """
 
         try:
-            return self.to_stream.write(self.from_stream.read(n))
+            read = self.from_stream.read(n)
+
+            if read is None or len(read) == 0:
+                self.eof = True
+                if self.propagate_close:
+                    self.to_stream.close()
+                return None
+
+            return self.to_stream.write(read)
         except OSError as e:
             if e.errno != errno.EPIPE:
                 raise e
 
+    def is_done(self):
+        """
+        Returns True if the read stream is done (either it's returned EOF or
+        the pump doesn't have wait_for_output set), and the write
+        side does not have pending bytes to send.
+        """
+
+        return (not self.wait_for_output or self.eof) and \
+                not (hasattr(self.to_stream, 'needs_write') and self.to_stream.needs_write())
+
     def __repr__(self):
         return "{cls}(from={from_stream}, to={to_stream})".format(
             cls=type(self).__name__,
