@@ -62,3 +62,9 @@ class BadFilename(BadSpecValue):
 class DeprecatedFeature(BadSpecValue):
     desc = "Feature is deprecated"
 
+class BadEnvironment(HarpoonError):
+    desc = "Something bad in the environment"
+
+class BadAmazon(HarpoonError):
+    desc = "Something wrong with amazon"
+
