@@ -10,20 +10,23 @@ Finally, the Runner is also responsible for starting and cleaning up interventio
 containers.
 """
 
-from __future__ import print_function
-
+from harpoon.errors import BadOption, BadImage, BadResult, UserQuit
 from harpoon.option_spec.harpoon_specs import HarpoonSpec
-from harpoon.errors import BadOption, BadImage, BadResult
+from harpoon import helpers as hp
 from harpoon.helpers import until
 
 from docker.errors import APIError as DockerAPIError
 from input_algorithms.spec_base import NotSpecified
 from input_algorithms.meta import Meta
 from contextlib import contextmanager
+from docker.utils import utils
 from harpoon import dockerpty
+from six.moves import input
+import docker.errors
 import logging
 import socket
 import uuid
+import time
 import os
 
 log = logging.getLogger("harpoon.ship.runner")
@@ -35,7 +38,7 @@ class Runner(object):
     ###   USAGE
     ########################
 
-    def run_container(self, conf, images, detach=False, started=None, dependency=False):
+    def run_container(self, conf, images, detach=False, started=None, dependency=False, tag=None):
         """Run this image and all dependency images"""
         if conf.container_id:
             return
@@ -46,16 +49,31 @@ class Runner(object):
             container_id = self.create_container(conf, detach, tty)
 
             conf.container_id = container_id
+
+            try:
+                self.wait_for_deps(conf, images)
+            except KeyboardInterrupt:
+                raise UserQuit()
+
             self.start_container(conf, tty=tty, detach=detach, is_dependency=dependency)
         finally:
             if not detach and not dependency:
                 self.stop_deps(conf, images)
-                self.stop_container(conf)
+                self.stop_container(conf, tag=tag)
+                self.delete_deps(conf, images)
 
     ########################
     ###   DEPS
     ########################
 
+    def delete_deps(self, conf, images):
+        """Delete any deleteable images"""
+        for dependency_name, _ in conf.dependency_images():
+            image = images[dependency_name]
+            if image.deleteable_image:
+                log.info("Removing un-needed image {0}".format(image.image_name))
+                conf.harpoon.docker_context.remove_image(image.image_name)
+
     def run_deps(self, conf, images):
         """Start containers for all our dependencies"""
         for dependency_name, detached in conf.dependency_images():
@@ -67,6 +85,7 @@ class Runner(object):
     def stop_deps(self, conf, images):
         """Stop the containers for all our dependencies"""
         for dependency, _ in conf.dependency_images():
+            self.stop_deps(images[dependency], images)
             try:
                 self.stop_container(images[dependency], fail_on_bad_exit=True, fail_reason="Failed to run dependency container")
             except BadImage:
@@ -74,6 +93,118 @@ class Runner(object):
             except Exception as error:
                 log.warning("Failed to stop dependency container\timage=%s\tdependency=%s\tcontainer_name=%s\terror=%s", conf.name, dependency, images[dependency].container_name, error)
 
+    def wait_for_deps(self, conf, images):
+        """Wait for all our dependencies"""
+        from harpoon.option_spec.image_objs import WaitCondition
+        ctxt = conf.harpoon.docker_context_maker()
+
+        waited = set()
+        last_attempt = {}
+        dependencies = set(dep for dep, _ in conf.dependency_images())
+
+        # Wait conditions come from dependency_options first
+        # Or if none specified there, they come from the image itself
+        wait_conditions = {}
+        for dependency in dependencies:
+            if conf.dependency_options is not NotSpecified and dependency in conf.dependency_options and conf.dependency_options[dependency].wait_condition is not NotSpecified:
+                wait_conditions[dependency] = conf.dependency_options[dependency].wait_condition
+            elif images[dependency].wait_condition is not NotSpecified:
+                wait_conditions[dependency] = images[dependency].wait_condition
+
+        if not wait_conditions:
+            return
+
+        start = time.time()
+        while True:
+            this_round = []
+            for dependency in dependencies:
+                if dependency in waited:
+                    continue
+
+                image = images[dependency]
+                if dependency in wait_conditions:
+                    done = self.wait_for_dep(ctxt, image, wait_conditions[dependency], start, last_attempt.get(dependency))
+                    this_round.append(done)
+                    if done is True:
+                        waited.add(dependency)
+                    elif done is False:
+                        last_attempt[dependency] = time.time()
+                    elif done is WaitCondition.Timedout:
+                        log.warning("Stopping dependency because it timedout waiting\tcontainer_id=%s", image.container_id)
+                        self.stop_container(image)
+                else:
+                    waited.add(dependency)
+
+            if set(this_round) != set([WaitCondition.KeepWaiting]):
+                if dependencies - waited == set():
+                    log.info("Finished waiting for dependencies")
+                    break
+                else:
+                    log.info("Still waiting for dependencies\twaiting_on=%s", list(dependencies-waited))
+
+                couldnt_wait = set()
+                container_ids = {}
+                for dependency in dependencies:
+                    if dependency in waited:
+                        continue
+
+                    image = images[dependency]
+                    if image.container_id is None:
+                        stopped = True
+                        if dependency not in container_ids:
+                            available = sorted([i for i in available if "/{0}".format(image.container_name) in i["Names"]], key=lambda i: i["Created"])
+                            if available:
+                                container_ids[dependency] = available[0]["Id"]
+                    else:
+                        if dependency not in container_ids:
+                            container_ids[dependency] = image.container_id
+                        stopped, _ = self.is_stopped(image, image.container_id)
+
+                    if stopped:
+                        couldnt_wait.add(dependency)
+
+                if couldnt_wait:
+                    for container in couldnt_wait:
+                        if container not in images or container not in container_ids:
+                            continue
+                        image = images[container]
+                        container_id = container_ids[container]
+                        container_name = image.container_name
+                        hp.write_to(conf.harpoon.stdout, "=================== Logs for failed container {0} ({1})\n".format(container_id, container_name))
+                        for line in conf.harpoon.docker_context.logs(container_id).split("\n"):
+                            hp.write_to(conf.harpoon.stdout, "{0}\n".format(line))
+                        hp.write_to(conf.harpoon.stdout, "------------------- End logs for failed container\n")
+                    raise BadImage("One or more of the dependencies stopped running whilst waiting for other dependencies", stopped=list(couldnt_wait))
+
+            time.sleep(0.1)
+
+    def wait_for_dep(self, ctxt, conf, wait_condition, start, last_attempt):
+        """Wait for this image"""
+        from harpoon.option_spec.image_objs import WaitCondition
+        conditions = list(wait_condition.conditions(start, last_attempt))
+        if conditions[0] in (WaitCondition.KeepWaiting, WaitCondition.Timedout):
+            return conditions[0]
+
+        log.info("Waiting for %s", conf.container_name)
+        for condition in conditions:
+            log.debug("Running condition\tcondition=%s", condition)
+            command = 'bash -c "{0}"'.format(condition)
+            try:
+                exec_id = ctxt.exec_create(conf.container_id, command, tty=False)
+            except DockerAPIError as error:
+                log.error("Failed to run condition\tcondition=%s\tdependency=%s\terror=%s", condition, conf.name, error)
+                return False
+
+            output = ctxt.exec_start(exec_id)
+            inspection = ctxt.exec_inspect(exec_id)
+            exit_code = inspection["ExitCode"]
+            if exit_code != 0:
+                log.error("Condition says no\tcondition=%s\toutput:\n\t%s", condition, "\n\t".join(line for line in output.split('\n')))
+                return False
+
+        log.info("Finished waiting for %s", conf.container_name)
+        return True
+
     ########################
     ###   CREATION
     ########################
@@ -85,11 +216,17 @@ class Runner(object):
         image_name = conf.image_name
         container_name = conf.container_name
 
-        env = dict(e.pair for e in conf.env)
+        with conf.assumed_role():
+            env = dict(e.pair for e in conf.env)
+
         ports = [port.host_port for port in conf.ports]
+        links = [link.pair for link in conf.links]
         binds = conf.volumes.binds
         command = conf.formatted_command
         volume_names = conf.volumes.volume_names
+        volumes_from = list(conf.volumes.share_with_names)
+        port_bindings = dict([port.pair for port in conf.ports])
+        no_tty_option = conf.no_tty_option
 
         uncreated = []
         for name in binds:
@@ -107,8 +244,44 @@ class Runner(object):
             log.info("\tUsing volumes\tvolumes=%s", volume_names)
         if env:
             log.info("\tUsing environment\tenv=%s", sorted(env.keys()))
+        if links:
+            log.info("\tLinks: %s", links)
         if ports:
             log.info("\tUsing ports\tports=%s", ports)
+        if port_bindings:
+            log.info("\tPort bindings: %s", ports)
+        if volumes_from:
+            log.info("\tVolumes from: %s", volumes_from)
+
+        host_config = conf.harpoon.docker_context.create_host_config(
+              links = links
+            , binds = binds
+            , volumes_from = volumes_from
+            , port_bindings = port_bindings
+
+            , devices = conf.devices
+            , lxc_conf = conf.lxc_conf
+            , privileged = conf.privileged
+            , restart_policy = conf.restart_policy
+
+            , dns = conf.network.dns
+            , dns_search = conf.network.dns_search
+            , extra_hosts = conf.network.extra_hosts
+            , network_mode = conf.network.network_mode
+            , publish_all_ports = conf.network.publish_all_ports
+
+            , cap_add = conf.cpu.cap_add
+            , cap_drop = conf.cpu.cap_drop
+            , mem_limit = conf.cpu.mem_limit
+            , memswap_limit = conf.cpu.memswap_limit
+
+            , ulimits = conf.ulimits
+            , read_only = conf.read_only_rootfs
+            , log_config = conf.log_config
+            , security_opt = conf.security_opt
+
+            , **conf.other_options.host_config
+            )
 
         container_id = conf.harpoon.docker_context.create_container(image_name
             , name=container_name
@@ -117,7 +290,7 @@ class Runner(object):
             , volumes=volume_names
             , environment=env
 
-            , tty = tty
+            , tty = False if no_tty_option else tty
             , user = conf.user
             , ports = ports
             , stdin_open = tty
@@ -128,9 +301,9 @@ class Runner(object):
             , network_disabled = conf.network.disabled
 
             , cpuset = conf.cpu.cpuset
-            , mem_limit = conf.cpu.mem_limit
             , cpu_shares = conf.cpu.cpu_shares
-            , memswap_limit = conf.cpu.memswap_limit
+
+            , host_config = host_config
 
             , **conf.other_options.create
             )
@@ -148,46 +321,20 @@ class Runner(object):
 
     def start_container(self, conf, tty=True, detach=False, is_dependency=False, no_intervention=False):
         """Start up a single container"""
-        binds = conf.volumes.binds
-        links = [link.pair for link in conf.links]
-        ports = dict([port.pair for port in conf.ports])
-        volumes_from = list(conf.volumes.share_with_names)
         container_id = conf.container_id
         container_name = conf.container_name
-
         log.info("Starting container %s (%s)", container_name, container_id)
-        if links:
-            log.info("\tLinks: %s", links)
-        if volumes_from:
-            log.info("\tVolumes from: %s", volumes_from)
-        if ports:
-            log.info("\tPort bindings: %s", ports)
 
-        conf.harpoon.docker_context.start(container_id
-            , links = links
-            , binds = binds
-            , port_bindings = ports
-            , volumes_from = volumes_from
-
-            , devices = conf.devices
-            , lxc_conf = conf.lxc_conf
-            , privileged = conf.privileged
-            , restart_policy = conf.restart_policy
-
-            , dns = conf.network.dns
-            , dns_search = conf.network.dns_search
-            , extra_hosts = conf.network.extra_hosts
-            , network_mode = conf.network.mode
-            , publish_all_ports = conf.network.publish_all_ports
-
-            , cap_add = conf.cpu.cap_add
-            , cap_drop = conf.cpu.cap_drop
-
-            , **conf.other_options.run
-            )
-
-        if not detach and not is_dependency:
-            self.start_tty(conf, interactive=tty)
+        try:
+            if not detach and not is_dependency:
+                self.start_tty(conf, interactive=tty, **conf.other_options.start)
+            else:
+                conf.harpoon.docker_context.start(container_id
+                    , **conf.other_options.start
+                    )
+        except docker.errors.APIError as error:
+            if str(error).startswith("404 Client Error: Not Found"):
+                log.error("Container died before we could even get to it...")
 
         inspection = None
         if not detach and not is_dependency:
@@ -196,14 +343,24 @@ class Runner(object):
         if inspection and not no_intervention:
             if not inspection["State"]["Running"] and inspection["State"]["ExitCode"] != 0:
                 self.stage_run_intervention(conf)
-                raise BadImage("Failed to run container", container_id=container_id, container_name=container_name)
+                raise BadImage("Failed to run container", container_id=container_id, container_name=container_name, reason="nonzero exit code after launch")
+
+        if not is_dependency and conf.harpoon.intervene_afterwards and not no_intervention:
+            self.stage_run_intervention(conf, just_do_it=True)
 
     def start_tty(self, conf, interactive):
         """Startup a tty"""
         try:
-            ctxt = conf.harpoon.docker_context
+            ctxt = conf.harpoon.docker_context_maker()
             container_id = conf.container_id
-            dockerpty.start(ctxt, container_id, interactive=interactive)
+
+            stdin = conf.harpoon.tty_stdin
+            stdout = conf.harpoon.tty_stdout
+            stderr = conf.harpoon.tty_stderr
+            if callable(stdin): stdin = stdin()
+            if callable(stdout): stdout = stdout()
+            if callable(stderr): stderr = stderr()
+            dockerpty.start(ctxt, container_id, interactive=interactive, stdout=stdout, stderr=stderr, stdin=stdin)
         except KeyboardInterrupt:
             pass
 
@@ -211,15 +368,11 @@ class Runner(object):
     ###   STOPPING
     ########################
 
-    def stop_container(self, conf, fail_on_bad_exit=False, fail_reason=None):
-        """Stop some container"""
+    def wait_till_stopped(self, conf, container_id, timeout=10, message=None, waiting=True):
+        """Wait till a container is stopped"""
         stopped = False
-        container_id = conf.container_id
-        if not container_id:
-            return
-
-        container_name = conf.container_name
-        for _ in until(timeout=10):
+        inspection = None
+        for _ in until(timeout=timeout, action=message):
             try:
                 inspection = conf.harpoon.docker_context.inspect_container(container_id)
                 if not isinstance(inspection, dict):
@@ -239,23 +392,46 @@ class Runner(object):
                 else:
                     break
 
-        if stopped:
+        if not inspection:
+            log.warning("Failed to inspect the container!")
+            stopped = True
+            exit_code = 1
+        else:
             exit_code = inspection["State"]["ExitCode"]
+        return stopped, exit_code
+
+    def is_stopped(self, *args, **kwargs):
+        """Return whether this container is stopped"""
+        kwargs["waiting"] = False
+        return self.wait_till_stopped(*args, **kwargs)
+
+    def stop_container(self, conf, fail_on_bad_exit=False, fail_reason=None, tag=None, remove_volumes=False):
+        """Stop some container"""
+        stopped = False
+        container_id = conf.container_id
+        if not container_id:
+            return
+
+        container_name = conf.container_name
+        stopped, exit_code = self.is_stopped(conf, container_id)
+
+        if stopped:
             if exit_code != 0 and fail_on_bad_exit:
-                if not self.harpoon.interactive:
+                if not conf.harpoon.interactive:
                     print_logs = True
                 else:
-                    print("!!!!")
-                    print("Container had already exited with a non zero exit code\tcontainer_name={0}\tcontainer_id={1}\texit_code={2}".format(container_name, container_id, exit_code))
-                    print("Do you want to see the logs from this container?")
-                    answer = raw_input("[y]: ")
+                    hp.write_to(conf.harpoon.stdout, "!!!!\n")
+                    hp.write_to(conf.harpoon.stdout, "Container had already exited with a non zero exit code\tcontainer_name={0}\tcontainer_id={1}\texit_code={2}\n".format(container_name, container_id, exit_code))
+                    hp.write_to(conf.harpoon.stdout, "Do you want to see the logs from this container?\n")
+                    conf.harpoon.stdout.flush()
+                    answer = input("[y]: ")
                     print_logs = not answer or answer.lower().startswith("y")
 
                 if print_logs:
-                    print("=================== Logs for failed container {0} ({1})".format(container_id, container_name))
+                    hp.write_to(conf.harpoon.stdout, "=================== Logs for failed container {0} ({1})\n".format(container_id, container_name))
                     for line in conf.harpoon.docker_context.logs(container_id).split("\n"):
-                        print(line)
-                    print("------------------- End logs for failed container")
+                        hp.write_to(conf.harpoon.stdout, "{0}\n".format(line))
+                    hp.write_to(conf.harpoon.stdout, "------------------- End logs for failed container\n")
                 fail_reason = fail_reason or "Failed to run container"
                 raise BadImage(fail_reason, container_id=container_id, container_name=container_name)
         else:
@@ -264,37 +440,38 @@ class Runner(object):
                 conf.harpoon.docker_context.kill(container_id, 9)
             except DockerAPIError:
                 pass
+            self.wait_till_stopped(conf, container_id, timeout=10, message="waiting for container to die\tcontainer_name={0}\tcontainer_id={1}".format(container_name, container_id))
 
-            for _ in until(timeout=10, action="waiting for container to die\tcontainer_name={0}\tcontainer_id={1}".format(container_name, container_id)):
+        if tag:
+            log.info("Tagging a container\tcontainer_id=%s\ttag=%s", container_id, tag)
+            new_id = conf.harpoon.docker_context.commit(container_id)["Id"]
+            conf["committed"] = new_id
+            if tag is not True:
+                conf.harpoon.docker_context.tag(new_id, repository=tag, tag="latest", force=True)
+
+        mounts = []
+        if remove_volumes:
+            for m in conf.harpoon.docker_context.inspect_container(container_id)["Mounts"]:
+                if "Name" in m:
+                    mounts.append(m['Name'])
+
+        if not conf.harpoon.no_cleanup:
+            log.info("Removing container %s:%s", container_name, container_id)
+            for _ in until(timeout=10, action="removing container\tcontainer_name={0}\tcontainer_id={1}".format(container_name, container_id)):
                 try:
-                    inspection = conf.harpoon.docker_context.inspect_container(container_id)
-                    if not inspection["State"]["Running"]:
-                        conf.container_id = None
-                        break
-                except socket.timeout:
-                    pass
-                except ValueError:
-                    log.warning("Failed to inspect the container\tcontainer_id=%s", container_id)
-                except DockerAPIError as error:
-                    if error.response.status_code != 404:
-                        raise
-                    else:
-                        break
-
-        log.info("Removing container %s:%s", container_name, container_id)
-        for _ in until(timeout=10, action="removing container\tcontainer_name={0}\tcontainer_id={1}".format(container_name, container_id)):
-            try:
-                conf.harpoon.docker_context.remove_container(container_id)
-                break
-            except socket.timeout:
-                break
-            except ValueError:
-                log.warning("Failed to remove container\tcontainer_id=%s", container_id)
-            except DockerAPIError as error:
-                if error.response.status_code != 404:
-                    raise
-                else:
+                    conf.harpoon.docker_context.remove_container(container_id)
                     break
+                except socket.timeout:
+                    break
+                except (ValueError, DockerAPIError) as error:
+                    log.warning("Failed to remove container\tcontainer_id=%s\terror=%s", container_id, error)
+
+        for mount in mounts:
+            try:
+                log.info("Cleaning up volume {0}".format(mount))
+                conf.harpoon.docker_context.remove_volume(mount)
+            except DockerAPIError as error:
+                log.warning("Failed to cleanup volume\tvolume=%s\terror=%s", mount, error)
 
     ########################
     ###   UTILITY
@@ -316,15 +493,19 @@ class Runner(object):
     ###   INTERVENTION
     ########################
 
-    def stage_run_intervention(self, conf):
+    def stage_run_intervention(self, conf, just_do_it=False):
         """Start an intervention!"""
         if not conf.harpoon.interactive or conf.harpoon.no_intervention:
             return
 
-        print("!!!!")
-        print("Failed to run the container!")
-        print("Do you want commit the container in it's current state and /bin/bash into it to debug?")
-        answer = raw_input("[y]: ")
+        if just_do_it:
+            answer = 'y'
+        else:
+            hp.write_to(conf.harpoon.stdout, "!!!!\n")
+            hp.write_to(conf.harpoon.stdout, "Failed to run the container!\n")
+            hp.write_to(conf.harpoon.stdout, "Do you want commit the container in it's current state and /bin/bash into it to debug?\n")
+            conf.harpoon.stdout.flush()
+            answer = input("[y]: ")
         if not answer or answer.lower().startswith("y"):
             with self.commit_and_run(conf.container_id, conf, command="/bin/bash"):
                 pass
@@ -356,10 +537,11 @@ class Runner(object):
             yield
             return
 
-        print("!!!!")
-        print("It would appear building the image failed")
-        print("Do you want to run /bin/bash where the build to help debug why it failed?")
-        answer = raw_input("[y]: ")
+        hp.write_to(conf.harpoon.stdout, "!!!!\n")
+        hp.write_to(conf.harpoon.stdout, "It would appear building the image failed\n")
+        hp.write_to(conf.harpoon.stdout, "Do you want to run /bin/bash where the build to help debug why it failed?\n")
+        conf.harpoon.stdout.flush()
+        answer = input("[y]: ")
         if answer and not answer.lower().startswith("y"):
             yield
             return
